#include <http.h>
#include <stdio.h>
#include "bldpath.h"
#include "debug.h"

#define STATUS_MSG_FILE_NOT_FOUND ("File Not Found")
#define STATUS_MSG_OK ("OK")

#define STATUS_BODY_FILE_NOT_FOUND ("<html><body><h1>File not found</h1></body></html>")


int _tmain(int argc, _TCHAR* argv[]) {

	ULONG ulRet;
	HTTPAPI_VERSION httpApiVersion = HTTPAPI_VERSION_2;

	//
	// Initialize HTTP API
	//
	
	ulRet = HttpInitialize (
		httpApiVersion,					// Version
		HTTP_INITIALIZE_SERVER,	// Flags
		NULL);									// Reserved

	//
	// Create Request Queue
	//

	HANDLE hReqQueue = NULL;

	ulRet = HttpCreateRequestQueue (
		httpApiVersion,
		NULL,
		NULL,
		0,
		&hReqQueue );

	if ( NO_ERROR != ulRet ) {
		printf ("HttpCreateRequestQueue Failed (%u)\n", ulRet);
		return 1;
	}

	//
	// Create Server Session
	//

	HTTP_SERVER_SESSION_ID httpServerSessionId = HTTP_NULL_ID;;

	ulRet = HttpCreateServerSession (
		httpApiVersion,
		&httpServerSessionId,
		0 );

	
	if ( NO_ERROR != ulRet ) {
		printf ("HttpCreateServerSession Failed (%u)\n", ulRet);
		goto cleanup;
	}

	//
	// Create URL Group
	//

	HTTP_URL_GROUP_ID httpUrlGroupId = HTTP_NULL_ID;

	ulRet = HttpCreateUrlGroup (
		httpServerSessionId,
		&httpUrlGroupId,
		0 );
	
	if ( NO_ERROR != ulRet ) {
		printf ("HttpCreateUrlGroup Failed (%u)\n", ulRet);
		goto cleanup;		
	}

	//
	// Add URL to the URL Group
	//

	ulRet = HttpAddUrlToUrlGroup (
		httpUrlGroupId,
		L"http://+:8080/",
		0,
		0 );

	if ( NO_ERROR != ulRet ) {
		printf ("HttpAddUrlToUrlGroup Failed (%u)\n", ulRet);
		goto cleanup;		
	}

	//
	// Set URL Group Property
	//

	HTTP_BINDING_INFO httpBindingInfo;
	HTTP_PROPERTY_FLAGS httpPropFlags;

	httpPropFlags.Present = 1;
	httpBindingInfo.Flags = httpPropFlags;
	httpBindingInfo.RequestQueueHandle = hReqQueue;
	
	ulRet = HttpSetUrlGroupProperty (
		httpUrlGroupId,
		HttpServerBindingProperty,
		&httpBindingInfo,
		sizeof( HTTP_BINDING_INFO )
	);

	if ( NO_ERROR != ulRet ) {
		printf ("HttpSetUrlGroupProperty Failed (%u)\n", ulRet);
		goto cleanup;		
	}

	wprintf (L"Start listening at http://+:8080/...\n");

	//
	// Send/Receive Loop
	//

	while (1) {

		HTTP_REQUEST_ID	HttpRequestID = HTTP_NULL_ID;
		PHTTP_REQUEST		pHttpRequest;
		BYTE				pHttpHeaderBuffer [2048]; // 2kb
		ULONG				cbHttpHeaderBuffer;
		ULONG				cbBytesReceived = 0;
		pHttpRequest			= (PHTTP_REQUEST) pHttpHeaderBuffer;
		cbHttpHeaderBuffer	= 2048;

		//
		// Read HTTP request synchronously
		//

		ulRet = HttpReceiveHttpRequest ( 
			hReqQueue, 
			HttpRequestID, 
			0, 
			pHttpRequest, 
			cbHttpHeaderBuffer, 
			&cbBytesReceived, 
			NULL);
	
		printf ( "HttpReceiveHttpRequest ... ");

		switch ( ulRet ) {
		case NO_ERROR:
			printf ( "NO_ERROR.\n");
			break;
		case ERROR_MORE_DATA:
			printf ("ERROR_MORE_DATA. cbBytesReceived=%u\n", cbBytesReceived);
			break;
		default:
			printf (" Failed %u.\n", ulRet );
		}

		//
		// Evaluate HTTP request
		//

		if ( HTTP_REQUEST_FLAG_MORE_ENTITY_BODY_EXISTS & pHttpRequest->Flags ) {
			
			// HTTP exists
			printf ("ENTITY_BODY_EXISTS!!\n");

			for (bool bContinue = true; bContinue; ) {

				BYTE		pHttpBodyBuffer [16];
				ULONG		cbHttpBodyBuffer = 16;
			
				cbBytesReceived = 0;

				ulRet = HttpReceiveRequestEntityBody (
					hReqQueue,
					pHttpRequest->RequestId,
					0, // Reserved
					pHttpBodyBuffer,
					cbHttpBodyBuffer,
					&cbBytesReceived,
					NULL);

				printf ("HttpReceiveRequestEntityBody returned ... cbBytesReceived=%u ", cbBytesReceived);
				switch ( ulRet ) {
				case NO_ERROR: // Received data successfully (but it's possible to reside data)
					printf ( "NO_ERROR\n" );

					// dump the data
					PrintHexDump ( cbBytesReceived, (PBYTE) pHttpBodyBuffer );
					
					break;
				case ERROR_HANDLE_EOF: // get all the data
					printf ( "ERROR_HANDLE_EOF\n" );
					bContinue = false;
					break;

				default:
					bContinue = false;
					printf ("%u\n", ulRet );
				}

			}

		}

		//
		// Send HTTP response synchronously
		//

		HTTP_RESPONSE			HttpResponse;
		::ZeroMemory ( &HttpResponse, sizeof (HTTP_RESPONSE) );
		HttpResponse.Version.MajorVersion = 1;
		HttpResponse.Version.MinorVersion = 0;
		HttpResponse.Headers.KnownHeaders[HttpHeaderContentType].pRawValue = "text/html";
		HttpResponse.Headers.KnownHeaders[HttpHeaderContentType].RawValueLength = strlen ("text/html");

		HTTP_DATA_CHUNK					HttpDataChunk;

		WCHAR		pszFileName[MAX_PATH];
		CHAR		pszContentType [64];
		ULONG		cbContentType = sizeof ( pszContentType );

		// Get file handle based on URL and root path.
		// If the returned handle is INVALID_HANDLE_VALUE, return 404 status code.
		HANDLE hFile = GetFileHandleFromURL ( 
			L"C:\\Inetpub\\wwwroot", 
			pHttpRequest->CookedUrl.pAbsPath, 
			pszFileName,											// out. File name
			MAX_PATH * sizeof (WCHAR));				// in.    buffer length of pszFileName in byte 

		if ( INVALID_HANDLE_VALUE == hFile ) { // 'File Not Found'
		
			// Server status
			HttpResponse.StatusCode = 404;
			HttpResponse.pReason = STATUS_MSG_FILE_NOT_FOUND;
			HttpResponse.ReasonLength = strlen ( STATUS_MSG_FILE_NOT_FOUND );

			// Data
			HttpDataChunk.DataChunkType		= HttpDataChunkFromMemory;
			HttpDataChunk.FromMemory.pBuffer = STATUS_BODY_FILE_NOT_FOUND;
			HttpDataChunk.FromMemory.BufferLength = strlen ( STATUS_BODY_FILE_NOT_FOUND );

			// Content-Type
			// (Default)
		
		}
		else {
			// Server status
			HttpResponse.StatusCode		= 200;
			HttpResponse.pReason			= STATUS_MSG_OK;
			HttpResponse.ReasonLength	= strlen ( STATUS_MSG_OK );
			
			// Data
			HttpDataChunk.DataChunkType		= HttpDataChunkFromFileHandle;
			HttpDataChunk.FromFileHandle.FileHandle = hFile;
			HttpDataChunk.FromFileHandle.ByteRange.StartingOffset.QuadPart = 0;
			HttpDataChunk.FromFileHandle.ByteRange.Length.QuadPart = HTTP_BYTE_RANGE_TO_EOF;

			// Content-Type
			if ( GetContentType ( pszFileName, pszContentType, &cbContentType ) ) {
				HttpResponse.Headers.KnownHeaders [HttpHeaderContentType].pRawValue = pszContentType;
				HttpResponse.Headers.KnownHeaders [HttpHeaderContentType].RawValueLength = cbContentType;
			}
		}

		HttpResponse.EntityChunkCount		= 1;
		HttpResponse.pEntityChunks			= &HttpDataChunk;
		ULONG	cbBytesSent = 0;
	
		ulRet = HttpSendHttpResponse ( 
			hReqQueue, 
			pHttpRequest->RequestId, 
			HTTP_SEND_RESPONSE_FLAG_DISCONNECT, 
			&HttpResponse, 
			NULL, 
			&cbBytesSent, 
			NULL, 
			0, 
			NULL, 
			NULL);

		printf ( "HttpSendHttpResponse (%u) cbBytesSent = %u\n", ulRet, cbBytesSent );
		
		// Close the file handle (if it's opened)
		if ( INVALID_HANDLE_VALUE != hFile ) {
			CloseHandle ( hFile );
		}

	} // loop

cleanup:

	//
	// clean up
	//

	if ( !HTTP_IS_NULL_ID ( &httpServerSessionId ) ) {

		HttpCloseServerSession ( httpServerSessionId );

	}
	
	// Close request queue
	
	if ( hReqQueue ) {
		
		HttpCloseRequestQueue ( hReqQueue );

	}

	// Terminate HTTP API
	HttpTerminate (HTTP_INITIALIZE_SERVER, NULL);

	return 0;
}

